/* gen.c

	Generate a linear code from Rz parse-tree
	see "linear-code.txt" for detail of the scheme  27 Nov 2010
*/

#include "compile.h"

#define head(e)		car(e)
#define tail(e) 	cdr(e)
#define arg1(e)		car(e)
#define arg2(e)		item2(e)
#define arg3(e)		item3(e)

int newsym = 100;			// global lable symbol, start at 100

//  RHS
void genatom(int a){
	int ty, val;
	ty = head(a);
	val = tail(a);
	switch(ty){
	case NUM: printf("  lit.%d\n",val); break;
	case STRING: printf("  str.%d\n",val); break;
	case GNAME: printf("  global.%d\n",getRef(val)); break;
	case LNAME: printf("  local.%d\n",val); break;
	}
}

void genbop(char *op, int e){
	gen(arg1(e));
	gen(arg2(e));
	printf("  %s\n",op);
}

void genuop(char *op, int e){
	gen(arg1(e));
	printf("  %s\n",op);
}
//  local var or global var with prefix
void genvar(char *prefix, int e){
	if( head(e) == GNAME )
		printf("  %sglobal.%d\n",prefix,getRef(tail(e)));
	else
	    printf("  %slocal.%d\n",prefix,tail(e));
}
// store global, put local, with prefix
void genstore(char *prefix, int e){
	if( head(e) == GNAME )
		printf("  %sst.%d\n",prefix,getRef(tail(e))); // global var
	else
	    printf("  %sput.%d\n",prefix,tail(e));		// local var
}
// generate new label
int newlabel(void){
	newsym++;
	return newsym;
}

void gen(int e){
	int op, v, ty;
	int idx, fs, e1, e2, a;

	if( e != NIL) {
		if( isatom(e) ) genatom(e);
		else{
			e1 = tail(e);			// argument list (arg1 arg2 ...)
			e = head(e);			// operator
			if(head(e) != OPER) seterror("expect operator");
			op = tail(e);
			switch(op){
			case tkIF:
				gen(arg1(e1));			// condition
				a = newlabel();
				printf("  jf L%d\n",a);	// jf exit
				gen(arg2(e1));			// action
				printf(":L%d\n",a);
				break;
			case tkELSE:
				gen(arg1(e1));			// condition
				a = newlabel();
				printf("  jf L%d\n",a);	// jf else
				gen(arg2(e1));			// action T
				v = newlabel();
				printf("  jmp L%d\n",v);// label exit
				printf(":L%d\n",a);
				gen(arg3(e1));
				printf(":L%d\n",v);
				break;
			case tkWHILE:
				a = newlabel();
				printf(":L%d\n",a);		// loop
				gen(arg1(e1));			// condition
				v = newlabel();
				printf("  jf L%d\n",v);	// jf exit
				gen(arg2(e1));			// body
				printf("  jmp L%d\n",a);// jmp loop
				printf(":L%d\n",v);		// exit
				break;
			case tkDO:
				while( e1 != NIL ){
					gen(head(e1));
					e1 = tail(e1);
				}
				break;
			case tkPLUS: genbop("add",e1); 	break;
			case tkMINUS: genbop("sub",e1); break;
			case tkSTAR: genbop("mul",e1); 	break;
			case tkSLASH: genbop("div",e1); break;
			case tkANDAND: genbop("and",e1); break;
			case tkOROR: genbop("or",e1); break;
			case tkEQEQ: genbop("eq",e1); break;
			case tkNE: genbop("ne",e1); break;
			case tkLT: genbop("lt",e1); break;
			case tkLE: genbop("le",e1); break;
			case tkGT: genbop("gt",e1); break;
			case tkGE: genbop("ge",e1); break;
			case tkUNOT: genuop("not",e1); break;
			case tkUMINUS: genbop("minus",e1); break;
			case tkCALL:
				idx = tail(head(e1));
				e1 = tail(e1);
				while( e1 != NIL ){		// gen all arg
					gen(head(e1));
					e1 = tail(e1);
				}
				printf("  call fun%d\n",idx);
				break;
			case tkRETURN: genuop("ret",e1); break;
			case tkFUN:
				idx = tail(arg1(e1));	// function
				a = getArg(idx);		// arity
				fs = getFs(idx);		// frame size
				printf(";; fun %s, arity %d, fs %d\n",getName(idx),a,fs);
				printf(":fun%d\n",idx);
				gen(arg2(e1));
				printf("  ret\n");
				break;
			case tkDEREF: genvar("*",head(e1)); break;	// RHS *
			case tkADS:	genvar("&",head(e1)); break;	// RHS &
			case tkVEC:
				gen(arg2(e1));			// idx
				genatom(arg1(e1));		// name
				printf("  ldx\n");		// load vector
				break;
			case tkEQ:					// assignment
				gen(arg2(e1));			// do RHS
				e1 = head(e1);			// LHS
				if ( isvec(e1) ){			// e1 is (vec name idx)
					gen(arg3(e1));			// index
					genatom(arg2(e1));		// name
					printf("  stx\n");		// store vector
				}
				else if( isderef(e1) )		// e1 is (* name)
					genstore("*",arg2(e1));
				else
					genstore("",e1);
				break;
			case tkPRINT:
				while( e1 != NIL ){
					e2 = head(e1);
					if(isatom(e2))
						genatom(e2);
					else
					    gen(e2);
					printf("  print\n");
					e1 = tail(e1);
				}
				break;
			default:
				seterror("unknown op\n");
			}
		}
	}
}

// end

